import json
import trello
import datetime
from collections import OrderedDict

import click
import prettytable

from todo.exceptions import GTDException
from todo.misc import Colors, get_banner


class Display:
    '''This class is responsible for displaying cards, lists, and other pieces of data from Trello in a visually appealing way.
    It replaces a polymorphic hierarchy that was a poor fit for this operation.
    Different functions are useful for displaying cards in JSON, a table, and in a set of pretty-printed lines.

    Needs:
    - json & text displays must be interchangeable for noninteractive commands
    - table display must have more columns / ability to select only certain columns
    - the "show lists/show tags" thing should have way more bits of metadata it can display. we should ideally have a method that just dumps the names of
      something to stdout
    - banner should have more ascii art options :D
    '''
    def __init__(self, color=True, primary_color=Colors.blue):
        self.color = color
        self.primary = primary_color

    def banner(self):
        '''Display an ASCII art banner for the beginning of program run'''
        print(get_banner(use_color=self.color))

    def show_raw(self, data, use_json=False):
        '''this shows random datastructures
        supports the following features
            show lists
            show tags
            show boards
        '''
        if use_json:
            print(json.dumps(self._force_json(data), sort_keys=True, indent=2))
        elif isinstance(data, list):
            for l in data:
                self.show_raw(l)
        elif isinstance(data, dict):
            for k, v in data.items():
                print(k, end=' ')
                self.show_raw(v)
        elif isinstance(data, trello.Board):
            print(data)
        elif isinstance(data, trello.List):
            print(data)
        elif isinstance(data, trello.Label):
            print(data)
        elif isinstance(data, trello.Card):
            print(data)
        else:
            print(data)

    def _force_json(self, for_json):
        '''force objects held in datastructures to be json-serializable by name only
        :param List|Label|Board|bytes|list|dict|datetime for_json: object to be encoded
        '''
        if isinstance(for_json, trello.List):
            return for_json.name
        elif isinstance(for_json, trello.Label):
            return for_json.name
        elif isinstance(for_json, trello.Board):
            return for_json.name
        elif isinstance(for_json, bytes):
            return for_json.decode('utf8')
        elif isinstance(for_json, list):
            return list(map(self._force_json, for_json))
        elif isinstance(for_json, dict):
            return {k: self._force_json(v) for k, v in for_json.items()}
        elif isinstance(for_json, datetime.datetime):
            return str(for_json)
        else:
            return for_json

    def show_cards(self, cards, use_json=False, tsv=False, table_fields=[], field_blacklist=[]):
        '''Display an iterable of cards all at once.
        Uses a pretty-printed table by default, but can also print JSON and tab-separated values (TSV).
        Supports the following cli commands:
            show cards
            grep

        :param list(trello.Card)|iterable(trello.Card) cards: cards to show
        :param bool use_json: display all metadata of these cards in JSON format
        :param bool tsv: display these cards using a tab-separated value format
        :param list table_fields: display only these fields (overrides field_blacklist)
        :param list field_blacklist: display all except these fields
        '''
        if use_json:
            sanitized_cards = list(map(
                lambda d: d.pop('client') and d,
                [c.__dict__.copy() for c in cards]
            ))
            tostr = self._force_json(sanitized_cards)
            print(json.dumps(tostr, sort_keys=True, indent=2))
        else:
            # TODO implement a custom sorting functions so the table can be sorted by multiple columns
            fields = OrderedDict()
            # This is done repetitively to establish column order
            fields['name'] = lambda c: c.name
            fields['list'] = lambda c: c.get_list().name
            fields['tags'] = lambda c: '\n'.join([l.name for l in c.list_labels]) if c.list_labels else ''
            fields['desc'] = lambda c: c.desc
            fields['due'] = lambda c: c.due or ''
            fields['last activity'] = lambda c: getattr(c, 'dateLastActivity')
            fields['board'] = lambda c: c.board.name
            fields['id'] = lambda c: getattr(c, 'id')
            fields['url'] = lambda c: getattr(c, 'shortUrl')
            table = prettytable.PrettyTable()
            table.field_names = fields.keys()
            table.align = 'l'
            if tsv:
                table.set_style(prettytable.PLAIN_COLUMNS)
            else:
                table.hrules = prettytable.FRAME
            with click.progressbar(list(cards), label='Fetching cards', width=0) as pg:
                for card in pg:
                    table.add_row([x(card) for x in fields.values()])
            try:
                table[0]
            except IndexError:
                click.secho('No cards match!', fg='red')
                raise GTDException(1)
            if table_fields:
                print(self.resize_and_get_table(table, table_fields))
            elif field_blacklist:
                f = set(fields.keys()) - set(field_blacklist)
                print(self.resize_and_get_table(table, list(f)))
            else:
                print(self.resize_and_get_table(table, fields.keys()))

    def resize_and_get_table(self, table, fields):
        '''Remove columns from the table until it fits in your terminal'''
        maxwidth = click.get_terminal_size()[0]
        possible = table.get_string(fields=fields, sortby='last activity')
        fset = set(fields)
        # Fields in increasing order of importance
        to_remove = ['desc', 'id', 'board', 'url', 'last activity', 'list']
        # Wait until we're under max width or until we can't discard more fields
        while len(possible.splitlines()[0]) >= maxwidth and to_remove:
            # Remove a field one at a time
            fset.remove(to_remove.pop(0))
            possible = table.get_string(fields=list(fset), sortby='last activity')
        return possible

    def show_card(self, card):
        '''Display only one card in a format that doesn't take up too much space or depend on external styling.
        Supports the following cli commands:
            review
            batch

        :param trello.Card card: card to display
        '''
        indent_print = lambda m, d: print('  {on}{name: <{fill}}{off}{val}'.format(name=m, val=d, fill='14', on=on, off=off))
        on = self.primary if self.color else ''
        off = Colors.reset if self.color else ''
        print('{on}Card{off}'.format(on=on, off=off), card.id)
        indent_print('Name:', card.name)
        indent_print('List:', '{0}'.format(card.get_list().name))
        if card.list_labels:
            indent_print('Tags:', ','.join([l.name for l in card.list_labels]))
        try:
            created = card.card_created_date
            indent_print('Created:', '{0} ({1})'.format(created, created.timestamp()))
            indent_print('Age:', datetime.datetime.now() - created)
        except IndexError:
            # this happens when the card is created by the repeating cards trello power-up
            indent_print('Created:', 'Repeating Creation Date')
        if card.get_attachments():
            indent_print('Attachments:', '')
            for a in card.get_attachments():
                print(' '*4 + a.name)
        if card.due:
            indent_print('Due:', card.due_date)
            try:
                diff = card.due_date - datetime.datetime.now(datetime.timezone.utc)
                if diff < datetime.timedelta(0):
                    display = Colors.red
                elif diff < datetime.timedelta(weeks=2):
                    display = Colors.yellow
                else:
                    display = Colors.green
                indent_print('Remaining:', '{0}{1}{2}'.format(display, diff, Colors.reset))
            except TypeError:
                # fucking datetime throws exceptions about bullshit
                pass
        if card.description:
            indent_print('Description', '')
            for line in card.description.splitlines():
                print(' '*4 + line)



import os
import yaml
import platform
from todo.exceptions import GTDException


class Configuration:
    '''hold global configuration for this application. This class has required
    arguments of the properties we need to connect to the Trello API and some
    other properties that modify global state during each run
    '''
    def __init__(self, api_key, api_secret, oauth_token, oauth_token_secret, **kwargs):
        self.api_key = api_key
        self.api_secret = api_secret
        self.oauth_token = oauth_token
        self.oauth_token_secret = oauth_token_secret
        self.board = kwargs.get('board', None)
        self.banner = kwargs.get('banner', True)
        self.color = kwargs.get('color', True)

    def __repr__(self):
        return '\n'.join([
            'GTD Configuration:',
            '  API key: ' + self.api_key,
            '  API secret: ' + self.api_secret,
            '  OAuth token: ' + self.oauth_token,
            '  OAuth secret: ' + self.oauth_token_secret,
            '  Primary board: '.format(self.board),
            '  Banner? '.format(self.banner),
            '  Use ANSI color? {0}'.format(self.color),
        ])

    def __str__(self):
        return repr(self)

    @staticmethod
    def suggest_config_location():
        '''Do some platform detection and suggest a place for the users' config file to go'''
        system = platform.system()
        if system == 'Windows':
            print('gtd.py support for Windows is rudimentary to none. Try to put your config file in $HOME/.gtd.yaml and run the script again')
            raise GTDException(0)
        elif system == 'Darwin':
            preferred_location = os.path.expanduser('~/Library/Application Support/gtd/gtd.yaml')
        elif system == 'Linux':
            preferred_location = os.path.expanduser('~/.config/gtd/gtd.yaml')
        else:
            preferred_location = os.path.expanduser('~/.gtd.yaml')
        return preferred_location

    @staticmethod
    def all_config_locations():
        return [os.path.expanduser(x) for x in [
            '~/.gtd.yaml',
            '~/.config/gtd/gtd.yaml',
            '~/Library/Application Support/gtd/gtd.yaml',
            '~/.local/etc/gtd.yaml',
            '~/.local/etc/gtd/gtd.yaml',
        ]]

    @staticmethod
    def find_config_file():
        # where to try finding the file in order
        for possible_loc in Configuration.all_config_locations():
            if os.path.isfile(possible_loc):
                return possible_loc
        # If we've gotten this far and did not find the configuration file, it does not exist
        raise GTDException(1)

    @staticmethod
    def from_file(filename=None):
        if filename is None:
            filename = Configuration.find_config_file()
        with open(filename, 'r') as config_yaml:
            file_config = yaml.safe_load(config_yaml)
        for prop in ['api_key', 'api_secret', 'oauth_token', 'oauth_token_secret']:
            if file_config.get(prop, None) is not None:
                # great!
                continue
            else:
                print('A required property {0} in your configuration was not found!'.format(prop))
                print('Check the file {0}'.format(filename))
                raise GTDException(1)
        return Configuration(
            file_config['api_key'],
            file_config['api_secret'],
            file_config['oauth_token'],
            file_config['oauth_token_secret'],
            board=file_config.get('board', None),
            color=file_config.get('color', True),
            banner=file_config.get('banner', True)
        )



#!/usr/bin/env python
'''entrypoint to gtd.py, written with click'''
import re
import os
import sys
import yaml
import click
import shutil
import requests
import readline  # noqa
import webbrowser
from requests_oauthlib import OAuth1Session
from requests_oauthlib.oauth1_session import TokenRequestDenied
from todo.input import prompt_for_confirmation, BoardTool, CardTool
from todo.display import Display
from todo.exceptions import GTDException
from todo.misc import Colors, DevNullRedirect, WORKFLOW_TEXT, get_banner, VALID_URL_REGEX
from todo.configuration import Configuration
from todo import __version__

pass_config = click.make_pass_decorator(Configuration)


def filtering_command(f):
    '''Add common options to a click function that will filter Trello cards'''
    f = click.option('-t', '--tags', default=None, help='Filter cards by this comma-separated list of tag names')(f)
    f = click.option('--no-tags', is_flag=True, default=False, help='Only show cards which have no tags')(f)
    f = click.option('-m', '--match', help='Filter cards to this regex on their title', default=None)(f)
    f = click.option('-l', '--listname', help='Only show cards from this list', default=None)(f)
    f = click.option('--attachments', is_flag=True, help='Only show cards which have attachments', default=None)(f)
    f = click.option('--has-due', is_flag=True, help='Only show cards which have due dates', default=None)(f)
    return f


def json_option(f):
    return click.option('-j', '--json', is_flag=True, default=False, help='Output as JSON')(f)


@click.group(context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(__version__)
@click.option('-b', '--board', default=None, help='Name of the board to work with for this session')
@click.option('--no-color', is_flag=True, default=False, help='Disable ANSI terminal color?')
@click.option('--no-banner', is_flag=True, default=False, help='Disable banner printing?')
@click.pass_context
def cli(ctx, board, no_color, no_banner):
    '''gtd.py'''
    try:
        config = Configuration.from_file()
    except GTDException:
        click.echo('Could not find a valid config file for gtd.')
        if click.confirm('Would you like to create it interactively?'):
            ctx.invoke(onboard)
            click.secho('Re-run your command', fg='green')
            raise GTDException(0)
        else:
            click.secho('Put your config file in one of the following locations:', fg='red')
            for l in Configuration.all_config_locations():
                print('  ' + l)
            raise
    if board is not None:
        config.board = board
    if no_color:
        config.color = False
    if no_banner:
        config.banner = False
    ctx.color = config.color
    ctx.obj = config


@cli.command()
@click.option('-w', '--workflow', is_flag=True, default=False, help='Show a Getting Things Done workflow')
@click.option('-b', '--banner', is_flag=True, default=False, help='Show a random banner')
def info(workflow, banner):
    '''Learn more about gtd.py'''
    if workflow:
        click.secho(WORKFLOW_TEXT, fg='yellow')
        raise GTDException(0)
    elif banner:
        print(get_banner())
    else:
        print('gtd.py version {c}{0}{r}'.format(__version__, c=Colors.green, r=Colors.reset))
        print('{c}https://github.com/delucks/gtd.py/{r}\nPRs welcome\n'.format(c=Colors.green, r=Colors.reset))


@cli.command()
@click.option('-e', '--edit', is_flag=True, help='Open $EDITOR on the configuration file')
def config(edit):
    '''Show/modify user configuration'''
    if edit:
        try:
            click.edit(filename=Configuration.find_config_file())
        except GTDException:
            # There is no configuration file
            click.secho("Could not find config file! Please run onboard if you haven't already", fg='red')
    else:
        print(Configuration.from_file())


# onboard {{{


@cli.command(short_help='Obtain Trello API credentials')
@click.option('-n', '--no-open', is_flag=True, default=False, help='Do not automatically open URLs in a web browser')
def onboard(no_open, output_path=None):
    '''Obtain Trello API credentials and put them into your config file.
    This is invoked automatically the first time you attempt to do an operation which requires authentication.
    The configuration file is put in an appropriate place for your operating system. If you want to change it later,
    you can use `gtd config -e` to open it in $EDITOR.
    '''
    output_file = output_path or Configuration.suggest_config_location()  # Use platform detection
    user_api_key_url = 'https://trello.com/app-key'
    request_token_url = 'https://trello.com/1/OAuthGetRequestToken'
    authorize_url = 'https://trello.com/1/OAuthAuthorizeToken'
    access_token_url = 'https://trello.com/1/OAuthGetAccessToken'
    # First, open the URL that allows the user to get an auth token. Tell them to copy both into the program
    click.echo('Welcome to gtd.py! To get started, open the following URL in your web browser:')
    click.echo('  ' + user_api_key_url)
    click.echo('When you arrive at that page, log in and copy the "Key" displayed in a text box.')
    if not no_open:
        with DevNullRedirect():
            webbrowser.open_new_tab(user_api_key_url)
    api_key = click.prompt('Please enter the value for "Key"', confirmation_prompt=True)
    click.echo('Now scroll to the bottom of the page and copy the "Secret" shown in a text box.')
    api_secret = click.prompt('Please enter the value for "Secret"', confirmation_prompt=True)
    # Then, work on getting OAuth credentials for the user so they are permanently authorized to use this program
    click.echo('We will now get the OAuth credentials necessary to use this program...')
    # The following code is cannibalized from trello.util.create_oauth_token from the py-trello project.
    # Rewriting because it does not support opening the auth URLs using webbrowser.open and since we're using
    # click, a lot of the input methods used in that script are simplistic compared to what's available to us.
    # Thank you to the original authors!
    '''Step 1: Get a request token. This is a temporary token that is used for
    having the user authorize an access token and to sign the request to obtain
    said access token.'''
    session = OAuth1Session(client_key=api_key, client_secret=api_secret)
    try:
        response = session.fetch_request_token(request_token_url)
    except TokenRequestDenied:
        click.secho('Invalid API key/secret provided: {0} / {1}'.format(api_key, api_secret), fg='red')
        sys.exit(1)
    resource_owner_key, resource_owner_secret = response.get('oauth_token'), response.get('oauth_token_secret')
    '''Step 2: Redirect to the provider. Since this is a CLI script we do not
    redirect. In a web application you would redirect the user to the URL
    below.'''
    user_confirmation_url = '{authorize_url}?oauth_token={oauth_token}&scope={scope}&expiration={expiration}&name={name}'.format(
        authorize_url=authorize_url,
        oauth_token=resource_owner_key,
        expiration='never',
        scope='read,write',
        name='gtd.py'
    )
    click.echo('Visit the following URL in your web browser to authorize gtd.py to access your account:')
    click.echo('  ' + user_confirmation_url)
    if not no_open:
        with DevNullRedirect():
            webbrowser.open_new_tab(user_confirmation_url)
    '''After the user has granted access to you, the consumer, the provider will
    redirect you to whatever URL you have told them to redirect to. You can
    usually define this in the oauth_callback argument as well.'''
    while not click.confirm('Have you authorized gtd.py?', default=False):
        pass
    oauth_verifier = click.prompt('What is the Verification code?').strip()
    '''Step 3: Once the consumer has redirected the user back to the oauth_callback
    URL you can request the access token the user has approved. You use the
    request token to sign this request. After this is done you throw away the
    request token and use the access token returned. You should store this
    access token somewhere safe, like a database, for future use.'''
    session = OAuth1Session(client_key=api_key, client_secret=api_secret,
                            resource_owner_key=resource_owner_key, resource_owner_secret=resource_owner_secret,
                            verifier=oauth_verifier)
    access_token = session.fetch_access_token(access_token_url)
    final_output_data = {
        'oauth_token': access_token['oauth_token'],
        'oauth_token_secret': access_token['oauth_token_secret'],
        'api_key': api_key,
        'api_secret': api_secret,
        'color': True,
        'banner': False
    }
    # Ensure we have a folder to put this in, if it's in a nested config location
    output_folder = os.path.dirname(output_file)
    if not os.path.isdir(output_folder):
        os.makedirs(output_folder)
        click.echo('Created folder {0} to hold your configuration'.format(output_folder))
    # Try to be safe about not destroying existing credentials
    if os.path.exists(output_file):
        if click.confirm('{0} exists already, would you like to back it up?'.format(output_file), default=True):
            shutil.move(output_file, output_file + '.backup')
        if not click.confirm('Overwrite the existing file?', default=False):
            return
    with open(output_file, 'w') as f:
        f.write(yaml.safe_dump(final_output_data, default_flow_style=False))
    click.echo('Credentials saved in "{0}"- you can now use gtd.py!'.format(output_file))
    click.echo('Use the "config" command to view or edit your configuration file')


# onboard }}}
# show {{{


@cli.group(short_help='Display cards, tags, or lists on this board')
def show():
    '''Display cards, tags, or lists on this board.'''
    pass


@show.command('lists')
@json_option
@click.option('-a', '--show-all', is_flag=True, default=False, help='Show open & archived lists')
@pass_config
def show_lists(config, json, show_all):
    '''Display all lists on this board'''
    _, board = BoardTool.start(config)
    display = Display(config.color)
    if config.banner and not json:
        display.banner()
    list_filter = 'all' if show_all else 'open'
    list_names = [l.name for l in board.get_lists(list_filter)]
    display.show_raw(list_names, use_json=json)


@show.command('tags')
@json_option
@pass_config
def show_tags(config, json):
    '''Display all tags on this board'''
    _, board = BoardTool.start(config)
    display = Display(config.color)
    if config.banner and not json:
        display.banner()
    tag_names = [t.name for t in board.get_labels()]
    display.show_raw(tag_names, use_json=json)


@show.command('cards')
@filtering_command
@json_option
@click.option('--tsv', is_flag=True, default=False, help='Output as tab-separated values')
@pass_config
def show_cards(config, json, tsv, tags, no_tags, match, listname, attachments, has_due):
    '''Display cards
    The show command prints a table of all the cards with fields that will fit on the terminal you're using.
    You can change this formatting by passing one of --tsv or --json, which will output as a tab-separated value sheet or JSON.
    This command along with the batch & review commands share a flexible argument scheme for getting card information.
    Mutually exclusive arguments include -t/--tags & --no-tags along with -j/--json & --tsv
    '''
    _, board = BoardTool.start(config)
    display = Display(config.color)
    if config.banner and not json:
        display.banner()
    cards = BoardTool.filter_cards(
        board,
        tags=tags,
        no_tags=no_tags,
        title_regex=match,
        list_regex=listname,
        has_attachments=attachments,
        has_due_date=has_due
    )
    display.show_cards(cards, use_json=json, tsv=tsv)


# show }}}
# add {{{


@cli.group()
def add():
    '''Add a new card, tag, or list'''
    pass


@add.command(short_help='Add a new card')
@click.argument('title', required=False)
@click.option('-m', '--message', help='Description for the new card')
@click.option('-e', '--edit', is_flag=True, help="Edit the card as soon as it's created")
@pass_config
def card(config, title, message, edit):
    '''Add a new card. If no title is provided, $EDITOR will be opened so you can write one.'''
    connection, board = BoardTool.start(config)
    inbox = BoardTool.get_inbox_list(connection, config)
    if not title:
        title = click.edit(require_save=True, text='Change this buffer to the title for your card')
        if title is None:  # No changes were made in $EDITOR
            click.secho('No title entered for the new card!', fg='red')
            raise GTDException(1)
        else:
            title = title.strip()
    returned = inbox.add_card(name=title, desc=message)
    if edit:
        display = Display(config.color)
        list_lookup = BoardTool.list_lookup(board)
        label_lookup = BoardTool.label_lookup(board)
        CardTool.smart_menu(returned, display.show_card, list_lookup, label_lookup, Colors.yellow)
    else:
        click.secho('Successfully added card {0}!'.format(returned), fg='green')


@add.command()
@click.argument('tagname')
@click.option('-c', '--color', help='color to create this tag with', default='black')
@pass_config
def tag(config, tagname, color):
    '''Add a new tag'''
    connection, board = BoardTool.start(config)
    label = board.add_label(tagname, color)
    click.secho('Successfully added tag {0}!'.format(label), color='green')


@add.command()
@click.argument('listname')
@pass_config
def list(config, listname):
    '''Add a new list'''
    connection, board = BoardTool.start(config)
    l = board.add_list(listname)
    click.secho('Successfully added list {0}!'.format(l), color='green')


@cli.command(short_help='egrep through titles of cards')
@click.argument('pattern', required=False)
@click.option('-i', '--insensitive', is_flag=True, help='Ignore case')
@click.option('-c', '--count', is_flag=True, help='Output the count of matching cards')
@click.option('-e', '--regexp', help='Specify multiple patterns to match against the titles of cards', multiple=True)
@pass_config
def grep(config, pattern, insensitive, count, regexp):
    '''egrep through titles of cards on this board. This command attemps to replicate a couple of grep flags
    faithfully, so if you're a power-user of grep this command will feel familiar.
    '''
    if not (pattern or regexp):
        click.secho('No pattern provided to grep: use either the argument or -e', fg='red')
        raise GTDException(1)
    # Merge together the different regex arguments
    final_pattern = '|'.join(regexp) if regexp else ''
    if pattern and final_pattern:
        final_pattern = final_pattern + '|' + pattern
    elif pattern:
        final_pattern = pattern
    flags = re.I if insensitive else 0
    connection, board = BoardTool.start(config)
    cards = BoardTool.filter_cards(
        board,
        title_regex=final_pattern,
        regex_flags=flags
    )
    if count:
        print(sum(1 for _ in cards))
        raise GTDException(0)
    display = Display(config.color)
    if config.banner:
        display.banner()
    display.show_cards(cards)


# add }}}
# batch {{{


@cli.group(short_help='Perform one action on many cards')
def batch():
    '''Perform one action on many cards at once. These commands are all interactive; they will prompt you before taking destructive action.
    If your combination of flags result in no cards matching, no output will be produced. You can tell this is not an error by the 0 return code.
    '''
    pass


@batch.command('move')
@filtering_command
@pass_config
def batch_move(config, tags, no_tags, match, listname, attachments, has_due):
    '''Change the list of each card selected'''
    connection, board = BoardTool.start(config)
    cards = BoardTool.filter_cards(
        board,
        tags=tags,
        no_tags=no_tags,
        title_regex=match,
        list_regex=listname,
        has_attachments=attachments,
        has_due_date=has_due
    )
    display = Display(config.color)
    if config.banner:
        display.banner()
    for card in cards:
        display.show_card(card)
        if prompt_for_confirmation('Want to move this one?', True):
            CardTool.move_to_list(card)


@batch.command('tag')
@filtering_command
@pass_config
def batch_tag(config, tags, no_tags, match, listname, attachments, has_due):
    '''Change tags on each card selected'''
    connection, board = BoardTool.start(config)
    cards = BoardTool.filter_cards(
        board,
        tags=tags,
        no_tags=no_tags,
        title_regex=match,
        list_regex=listname,
        has_attachments=attachments,
        has_due_date=has_due
    )
    display = Display(config.color)
    if config.banner:
        display.banner()
    for card in cards:
        display.show_card(card)
        CardTool.add_labels(card)


@batch.command('due')
@filtering_command
@pass_config
def batch_due(config, tags, no_tags, match, listname, attachments, has_due):
    '''Set due date for all cards selected'''
    connection, board = BoardTool.start(config)
    cards = BoardTool.filter_cards(
        board,
        tags=tags,
        no_tags=no_tags,
        title_regex=match,
        list_regex=listname,
        has_attachments=attachments,
        has_due_date=has_due
    )
    display = Display(config.color)
    if config.banner:
        display.banner()
    for card in cards:
        display.show_card(card)
        if prompt_for_confirmation('Set due date?'):
            CardTool.set_due_date(card)


@batch.command('attach')
@pass_config
def batch_attach(config):
    '''Extract HTTP links from card titles'''
    connection, board = BoardTool.start(config)
    cards = BoardTool.filter_cards(board, title_regex=VALID_URL_REGEX)
    display = Display(config.color)
    if config.banner:
        display.banner()
    for card in cards:
        display.show_card(card)
        if prompt_for_confirmation('Attach title?', True):
            CardTool.title_to_link(card)


@batch.command('delete')
@filtering_command
@pass_config
def batch_delete(config, tags, no_tags, match, listname, attachments, has_due):
    '''Interactively delete all cards selected'''
    connection, board = BoardTool.start(config)
    cards = BoardTool.filter_cards(
        board,
        tags=tags,
        no_tags=no_tags,
        title_regex=match,
        list_regex=listname,
        has_attachments=attachments,
        has_due_date=has_due
    )
    display = Display(config.color)
    if config.banner:
        display.banner()
    for card in cards:
        display.show_card(card)
        if prompt_for_confirmation('Should we delete this card?'):
            card.delete()
            click.secho('Card deleted!', fg='red')


# batch }}}


@cli.command(short_help='Use a smart shell-like menu')
@filtering_command
@pass_config
def review(config, tags, no_tags, match, listname, attachments, has_due):
    '''show a smart, command-line based menu for each card selected.
    This menu will prompt you to add tags to untagged cards, to attach the title
    of cards which have a link in the title, and gives you all the other functionality combined.
    '''
    connection, board = BoardTool.start(config)
    cards = BoardTool.filter_cards(
        board,
        tags=tags,
        no_tags=no_tags,
        title_regex=match,
        list_regex=listname,
        has_attachments=attachments,
        has_due_date=has_due
    )
    list_lookup = BoardTool.list_lookup(board)
    label_lookup = BoardTool.label_lookup(board)
    display = Display(config.color)
    if config.banner:
        display.banner()
    for card in cards:
        CardTool.smart_menu(card, display.show_card, list_lookup, label_lookup, Colors.yellow)
    click.echo('All done, have a great day!')


def main():
    try:
        cli()
    except requests.exceptions.ConnectionError:
        click.secho('[FATAL] Connection lost to the Trello API!', fg='red')
        sys.exit(1)
    except GTDException as e:
        if e.errno == 0:
            sys.exit(0)
        else:
            sys.exit(1)


if __name__ == '__main__':
    main()



import os
import requests
from todo import __version__, __author__
try:
    from random import choice
except OSError:
    # Some platforms (cough cough armv7, mips) fail when importing random due to lack of hardware support
    choice = lambda n: n.pop()


VALID_URL_REGEX = 'https?://.*\.'


class Colors:
    esc = '\033'
    black = esc + '[0;30m'
    red = esc + '[0;31m'
    green = esc + '[0;32m'
    yellow = esc + '[0;33m'
    blue = esc + '[0;34m'
    purple = esc + '[0;35m'
    cyan = esc + '[0;36m'
    white = esc + '[0;37m'
    reset = esc + '[0m'

    @staticmethod
    def all_colors():
        return [Colors.red, Colors.green, Colors.yellow, Colors.blue, Colors.purple, Colors.cyan]


def get_title_of_webpage(url):
    headers = {'User-Agent': 'gtd.py version ' + __version__}
    try:
        resp = requests.get(url, headers=headers)
        as_text = resp.text
        return as_text[as_text.find('<title>') + 7:as_text.find('</title>')]
    except requests.exceptions.ConnectionError:
        return None


class DevNullRedirect:
    '''Temporarily eat stdout/stderr to allow no output.
    This is used to suppress browser messages in webbrowser.open'''
    def __enter__(self):
        self.old_stderr = os.dup(2)
        self.old_stdout = os.dup(1)
        os.close(2)
        os.close(1)
        os.open(os.devnull, os.O_RDWR)

    def __exit__(self, exc_type, exc_value, traceback):
        os.dup2(self.old_stderr, 2)
        os.dup2(self.old_stdout, 1)


WORKFLOW_TEXT = (
    '1. Collect absolutely everything that can take your attention into "Inbound"\n'
    '2. Filter:\n'
    '    Nonactionable -> Static Reference or Delete\n'
    '    Takes < 2 minutes -> Do now, then Delete\n'
    '    Not your responsibility -> "Holding" or "Blocked" with follow-up\n'
    '    Something to communicate -> messaging lists\n'
    '    Your responsibility -> Your lists\n'
    '3. Write "final" state of each task and "next" state of each task\n'
    '4. Categorize inbound items into lists based on action type required (call x, talk to x, meet x...)\n'
    '5. Reviews:\n'
    '    Daily -> Go through "Inbound" and "Doing"\n'
    '    Weekly -> Additionally, go through "Holding", "Blocked", and messaging lists\n'
    '6. Do\n'
    '\n'
    'The goal is to get everything except the current task out of your head\n'
    'and into a trusted system external to your mind.'
)


def get_banner(version=__version__, author=__author__, use_color=True):
    '''Hold a buncha poorly done ASCII banners and display one at random!'''
    if use_color:
        on = choice(Colors.all_colors())
        off = Colors.reset
    else:
        on = off = ''
    b1 = (
        ' __|_ _| ._     version {on}{0}{off}\n'
        '(_||_(_|{on}o{off}|_)\/  by {on}{1}{off}\n'
        ' _|      |  /\n'
    ).format(version, author, on=on, off=off)
    b2 = (
        '  ___  ____  ____    ____  _  _\n'
        ' / __)(_  _)(  _ \  (  _ \( \/ )\n'
        '( (_-.  )(   )(_) )  )___/ \  /\n'
        ' \___/ (__) (____/{on}(){off}(__)   (__)\n'
        '   version {on}{0}{off}\n'
    ).format(version, on=on, off=off)
    b3 = (
        '        __      __   version {on}{0}{off}\n'
        '.-----.|  |_.--|  |  .-----.--.--.\n'
        '|  _  ||   _|  _  |{on}__{off}|  _  |  |  |\n'
        '|___  ||____|_____{on}|__|{off}   __|___  |\n'
        '|_____|              |__|  |_____|\n'
    ).format(version, on=on, off=off)
    b4 = (
        '         __      .___  version {on}{0}{off}\n'
        '   _____/  |_  __| _/______ ___.__.\n'
        '  / ___\   __\/ __ | \____ <   |  |\n'
        ' / /_/  >  | / /_/ | |  |_> >___  |\n'
        ' \___  /|__| \____ |{on}/\{off}   __// ____|\n'
        '/_____/           \/{on}\/{off}__|   \/\n'
    ).format(version, on=on, off=off)
    # A joke, is funny
    b5 = '67 74 64 {on}2e{off} 70 79\n76 65 72 73 69 6f 6e {on}{0}{off}\n'.format(version, on=on, off=off)
    b6 = '--. - -.. {on}.-.-.-{off} .--. -.--\n...- . .-. ... .. --- -. {on}{0}{off}\n'.format(version, on=on, off=off)
    b7 = (
        '     _     _\n'
        ' ___| |_ _| |  ___ _ _\n'
        '| . |  _| . |{on}_{off}| . | | |\n'
        '|_  |_| |___{on}|_|{off}  _|_  |\n'
        '|___|         |_| |___|\n'
        '     version {on}{0}{off}\n'
    ).format(version, on=on, off=off)
    banner_choices = [b1, b2, b3, b4, b5, b6, b7]
    return choice(banner_choices)
